from abc import ABC, abstractmethod
from typing import Any

from bepatient.waiter_src.checkers.checker import Checker
from bepatient.waiter_src.exceptions.executor_exceptions import ExecutorIsNotReady


class Executor(ABC):
    """An abstract base class for defining an executor that can be waited for."""

    def __init__(self):
        """Cursor should have cursor_factory that returns dict object"""
        self._checkers: list[Checker] = []
        self._failed_checkers: list[Checker] = []
        self._result: Any = None
        self._input: str | None = None

    def add_checker(self, checker: Checker):
        """Adds checker function to the list of checkers."""
        self._checkers.append(checker)
        return self

    @abstractmethod
    def is_condition_met(self) -> bool:
        """Check whether the condition has been met.

        Returns:
            bool: True if the condition has been met, False otherwise."""

    def get_result(self) -> Any:
        """Returns the result of performed actions."""
        if self._result is not None:
            return self._result
        raise ExecutorIsNotReady()

    def error_message(self) -> str:
        """Return a detailed error message if the condition has not been met."""
        if self._result is not None and len(self._failed_checkers) > 0:
            checkers = ", ".join([str(checker) for checker in self._failed_checkers])
            return (
                "The condition has not been met!"
                f" | Failed checkers: ({checkers})"
                f" | {self._input}"
            )
        if self._result is not None and len(self._failed_checkers) == 0:
            return "All conditions have been met."
        raise ExecutorIsNotReady()



"""A library facilitating work with asynchronous APIs"""
import logging
from logging import NullHandler

from .api import (
    RequestsWaiter,
    dict_differences,
    to_curl,
    wait_for_value_in_request,
    wait_for_values_in_request,
)
from .waiter_src.checkers import CHECKERS
from .waiter_src.checkers.checker import Checker
from .waiter_src.comparators import COMPARATORS

__version__ = "0.10.0"
__all__ = [
    "Checker",
    "CHECKERS",
    "COMPARATORS",
    "dict_differences",
    "RequestsWaiter",
    "to_curl",
    "wait_for_values_in_request",
    "wait_for_value_in_request",
]

logging.getLogger(__name__).addHandler(NullHandler())



import logging
from json import JSONDecodeError
from typing import Any, Callable

from dictor import dictor
from requests import Response

from .checker import Checker

log = logging.getLogger(__name__)


class StatusCodeChecker(Checker):
    def prepare_data(self, data: Response, run_uuid: str | None = None) -> int:
        """Prepare the response status code for comparison.

        Args:
            data (Response): response containing the status code.
            run_uuid (str | None, optional): unique run identifier. Defaults to None.

        Returns:
            int: prepared status code for comparison."""
        status_code = data.status_code
        log.info(
            "Check uuid: %s | Response status code: %s | Response content: %s",
            run_uuid,
            status_code,
            data.content,
        )
        return status_code


class JsonChecker(Checker):
    """A checker that compares a value in a JSON response against an expected value.

    Args:
        comparer (Callable): A function that performs the comparison.
        expected_value (Any): The expected value to compare against.
        dict_path (str, optional): dot-separated path to the value in the response data.
            Defaults to None.
        search_query (str, optional): A search query to find the value in the response
            data. Defaults to None.
        dictor_fallback (str, optional): A default value to return if the value at the
            specified path is not found using `dictor`. Defaults to None.
        ignore_case (bool, optional): If set, upper/lower-case keys in dict_path are
            treated the same. Defaults to False.

    Example:
        To check if a specific field "status" in a JSON response equals 200:

        ```
            checker = JsonChecker(lambda a, b: a == b, 200, dict_path="status")
            assert checker.check(response) is True
        ```"""

    def __init__(
        self,
        comparer: Callable[[Any, Any], bool],
        expected_value: Any,
        dict_path: str | None = None,
        search_query: str | None = None,
        dictor_fallback: str | None = None,
        ignore_case: bool = False,
    ):
        super().__init__(comparer, expected_value)
        self.path = dict_path
        self.search_query = search_query
        self.dictor_fallback = dictor_fallback
        self.ignore_case = ignore_case

    @staticmethod
    def parse_response(
        data: Response, run_uuid: str | None = None
    ) -> dict[str, Any] | list[Any]:
        """Parse the response content as JSON for comparison.

        Args:
            data (Response): response containing the JSON data.
            run_uuid (str | None): unique run identifier. Defaults to None.

        Returns:
            dict[str, Any] | list[Any]: The parsed JSON response data for comparison."""
        log.debug("Check uuid: %s | Response content: %s", run_uuid, data.content)
        return data.json()

    def prepare_data(self, data: Response, run_uuid: str | None = None) -> Any:
        """Prepare the response data for comparison.

        Args:
            data (Response): The response containing the data.
            run_uuid (str | None): The unique run identifier. Defaults to None.

        Returns:
            Any: The prepared data for comparison."""
        try:
            dictor_data = dictor(
                data=self.parse_response(data, run_uuid),
                path=self.path,
                search=self.search_query,
                default=self.dictor_fallback,
                ignorecase=self.ignore_case,
            )
            log.info(
                "Check uuid: %s | Dictor path: %s"
                " | Dictor search: %s | Dictor data: %s",
                run_uuid,
                self.path,
                self.search_query,
                dictor_data,
            )
            return dictor_data
        except (TypeError, JSONDecodeError):
            log.exception(
                "Check uuid: %s | Expected: %s | Headers: %s | Content %s",
                run_uuid,
                self.expected_value,
                data.headers,
                data.content,
            )
        return None


class HeadersChecker(JsonChecker):
    """A checker that compares response headers against expected values.

    Example:
        To check if the "Content-Type" header in a response is "application/json":
        ```
            checker = HeadersChecker(
                lambda a, b: a == b, "application/json", dict_path="Content-Type"
            )
            assert checker.check(response) is True
        ```
    """

    @staticmethod
    def parse_response(data: Response, run_uuid: str | None = None) -> dict[str, str]:
        """Parse the response headers for comparison.

        Args:
            data (Response): The response containing the headers.
            run_uuid (str | None): The unique run identifier. Defaults to None.

        Returns:
            dict[str, str]: The parsed response headers for comparison."""
        headers = dict(data.headers)
        log.info("Check uuid: %s | Response headers: %s", run_uuid, headers)
        return headers



import json
from typing import Any

import pytest
from _pytest.logging import LogCaptureFixture
from pytest_mock import MockerFixture
from requests import PreparedRequest, Request, Response, Session
from responses import RequestsMock

from bepatient import (
    RequestsWaiter,
    dict_differences,
    to_curl,
    wait_for_value_in_request,
    wait_for_values_in_request,
)
from bepatient.waiter_src.checkers.response_checkers import HeadersChecker
from bepatient.waiter_src.comparators import is_equal
from bepatient.waiter_src.exceptions.waiter_exceptions import WaiterConditionWasNotMet


class TestRequestsWaiter:
    def test_init(self, prepared_request: PreparedRequest, session_mock: Session):
        executor = RequestsWaiter(
            request=prepared_request, status_code=201, session=session_mock
        ).executor

        assert executor.request == prepared_request
        assert executor.session is session_mock

        checker = executor._status_code_checker
        assert checker.expected_value == 201

    def test_add_checker(self, prepared_request: PreparedRequest):
        waiter = RequestsWaiter(request=prepared_request)
        executor = waiter.add_checker(
            expected_value="TEST",
            comparer="is_equal",
            checker="headers_checker",
            dict_path="dict",
            search_query="search",
        ).executor

        checker = HeadersChecker(
            comparer=is_equal,
            expected_value="TEST",
            dict_path="dict",
            search_query="search",
        )

        assert executor._checkers[0].__dict__ == checker.__dict__

    def test_add_custom_checker(self, prepared_request: PreparedRequest):
        checker = HeadersChecker(isinstance, dict)
        waiter = RequestsWaiter(request=prepared_request)
        executor = waiter.add_custom_checker(checker).executor
        w_checker = executor._checkers[0]

        assert w_checker is checker

    def test_happy_path_prepared_request(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        example_response: Response,
    ):
        waiter = RequestsWaiter(request=prepared_request, session=session_mock)
        waiter.add_checker(expected_value=True, comparer="is_equal", dict_path="ok")
        response = waiter.run(retries=1).get_result()

        assert response == example_response

    def test_happy_path_request(
        self,
        mocked_responses: RequestsMock,
        request_object: Request,
        example_dict_content: dict[str, Any],
        mocker: MockerFixture,
        session_object: Session,
        caplog: LogCaptureFixture,
    ):
        mocked_responses.get(request_object.url, status=404)
        mocked_responses.get(request_object.url, status=200)
        mocked_responses.get(
            request_object.url,
            status=200,
            body=json.dumps(example_dict_content).encode("utf-8"),
        )
        mocker.patch("uuid.uuid4", side_effect=[1, 2, 3, 4, 5])
        logs = [
            (
                "bepatient.waiter_src.waiter",
                20,
                "Checking whether the condition has been met. The 1 approach",
            ),
            (
                "bepatient.waiter_src.executors.requests_executor",
                20,
                "Sent: curl -X GET -H 'Content-Type: application/json' -H"
                " 'Accept-Language: en-US,en;' -H 'Host: webludus.pl' -H 'User-Agent:"
                " Mozilla/5.0 (Windows NT 10.0; rv:120.0) Gecko/20100101' -H 'task:"
                " test' -H 'Cookie: pytest=fixture; user-token=abc-123'"
                " https://webludus.pl/",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: 1 | Checker: StatusCodeChecker | Comparer: is_equal"
                " | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: 1 | Response status code: 404 | Response content: b''",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: 1 | Condition not met | Checker: StatusCodeChecker"
                " | Comparer: is_equal | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.waiter",
                20,
                "The condition has not been met. Waiting time: 1",
            ),
            (
                "bepatient.waiter_src.waiter",
                20,
                "Checking whether the condition has been met. The 2 approach",
            ),
            (
                "bepatient.waiter_src.executors.requests_executor",
                20,
                "Sent: curl -X GET -H 'Content-Type: application/json' -H"
                " 'Accept-Language: en-US,en;' -H 'Host: webludus.pl' -H 'User-Agent:"
                " Mozilla/5.0 (Windows NT 10.0; rv:120.0) Gecko/20100101' -H 'task:"
                " test' -H 'Cookie: pytest=fixture; user-token=abc-123'"
                " https://webludus.pl/",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: 2 | Checker: StatusCodeChecker | Comparer: is_equal"
                " | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: 2 | Response status code: 200 | Response content: b''",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                10,
                "Check success! | uuid: 2 | Checker: StatusCodeChecker"
                " | Comparer: is_equal | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: 3 | Checker: JsonChecker | Comparer: is_equal"
                " | Dictor_fallback: None | Expected_value: Jack | Ignore_case: False"
                " | Path: name | Search_query: None | Data: Jack",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                10,
                "Check uuid: 3 | Response content: b''",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                40,
                "Check uuid: 3 | Expected: Jack | Headers:"
                " {'Content-Type': 'text/plain'} | Content b''",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: 3 | Condition not met | Checker: JsonChecker"
                " | Comparer: is_equal | Dictor_fallback: None | Expected_value: Jack"
                " | Ignore_case: False | Path: name | Search_query: None | Data: Jack",
            ),
            (
                "bepatient.waiter_src.waiter",
                20,
                "The condition has not been met. Waiting time: 1",
            ),
            (
                "bepatient.waiter_src.waiter",
                20,
                "Checking whether the condition has been met. The 3 approach",
            ),
            (
                "bepatient.waiter_src.executors.requests_executor",
                20,
                "Sent: curl -X GET -H 'Content-Type: application/json' -H"
                " 'Accept-Language: en-US,en;' -H 'Host: webludus.pl' -H 'User-Agent:"
                " Mozilla/5.0 (Windows NT 10.0; rv:120.0) Gecko/20100101' -H 'task:"
                " test' -H 'Cookie: pytest=fixture; user-token=abc-123'"
                " https://webludus.pl/",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: 4 | Checker: StatusCodeChecker | Comparer: is_equal"
                " | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: 4 | Response status code: 200 | Response content: "
                'b\'{"list_of_dicts": [{"name": "John", "age": 30}, {"name": "Mike",'
                ' "age": 15}], "ok": true, "some_number": 123, "list": ["1", "2", "3"],'
                ' "none": null, "empty": "", "false": false, "name": "Jack",'
                ' "City": "Cracow"}\'',
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                10,
                "Check success! | uuid: 4 | Checker: StatusCodeChecker"
                " | Comparer: is_equal | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: 5 | Checker: JsonChecker | Comparer: is_equal"
                " | Dictor_fallback: None | Expected_value: Jack | Ignore_case: False"
                " | Path: name | Search_query: None | Data: Jack",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                10,
                'Check uuid: 5 | Response content: b\'{"list_of_dicts": [{"name":'
                ' "John", "age": 30}, {"name": "Mike", "age": 15}], "ok": true,'
                ' "some_number": 123, "list": ["1", "2", "3"], "none": null, "empty":'
                ' "", "false": false, "name": "Jack", "City": "Cracow"}\'',
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: 5 | Dictor path: name | Dictor search: None"
                " | Dictor data: Jack",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                10,
                "Check success! | uuid: 5 | Checker: JsonChecker | Comparer: is_equal"
                " | Dictor_fallback: None | Expected_value: Jack | Ignore_case: False"
                " | Path: name | Search_query: None | Data: Jack",
            ),
            ("bepatient.waiter_src.waiter", 20, "Condition met!"),
        ]

        waiter = RequestsWaiter(request=request_object, session=session_object)
        waiter.add_checker(expected_value="Jack", comparer="is_equal", dict_path="name")
        res_json = waiter.run(retries=3).get_result().json()

        mocked_responses.assert_call_count(url=request_object.url, count=3)
        assert res_json == example_dict_content
        assert caplog.record_tuples == logs

    def test_happy_path(
        self,
        mocked_responses: RequestsMock,
        example_dict_content: dict[str, Any],
        prepared_request: PreparedRequest,
        session_object: Session,
    ):
        res = Response()
        res.status_code = 404
        res.request = prepared_request

        mocked_responses.get(
            res.request.url, status=200, headers={"Content-Type": "application/json"}
        )
        mocked_responses.get(
            res.request.url,
            status=200,
            headers={"Content-Type": "application/json", "Server": "WebLudus.pl"},
            body=json.dumps(example_dict_content).encode("utf-8"),
        )

        waiter = RequestsWaiter(request=res, session=session_object)
        waiter.add_checker(
            expected_value="WebLudus.pl",
            comparer="is_equal",
            checker="headers_checker",
            dict_path="Server",
        )
        error_message = (
            "The condition has not been met!"
            " | Failed checkers: (Checker: HeadersChecker | Comparer: is_equal"
            " | Dictor_fallback: None | Expected_value: WebLudus.pl"
            " | Ignore_case: False | Path: Server | Search_query: None | Data: None)"
            " | curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123;"
            " pytest=fixture' -H 'Content-Type: application/json' -H 'Accept-Language:"
            " en-US,en;' -H 'Host: webludus.pl' -H 'User-Agent: Mozilla/5.0"
            " (Windows NT 10.0; rv:120.0) Gecko/20100101' https://webludus.pl/"
        )
        with pytest.raises(WaiterConditionWasNotMet, match=error_message):
            waiter.run(retries=2)
        res_json = waiter.run(retries=1).get_result().json()

        mocked_responses.assert_call_count(url=res.request.url, count=2)  # type: ignore
        assert res_json == example_dict_content

    def test_happy_path_with_custom_checker(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        example_response: Response,
    ):
        waiter = RequestsWaiter(request=prepared_request, session=session_mock)
        checker = HeadersChecker(
            comparer=isinstance, expected_value=str, dict_path="Server"
        )
        waiter.add_custom_checker(checker)
        response = waiter.run(retries=1).get_result()

        assert response == example_response
        assert response.headers["Content-Type"] == "application/json"

    def test_condition_not_met_raise_error(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        example_response: Response,
        error_msg: str,
    ):
        waiter = RequestsWaiter(request=prepared_request, session=session_mock)
        waiter.add_checker(expected_value=False, comparer="is_equal", dict_path="ok")
        with pytest.raises(WaiterConditionWasNotMet, match=error_msg):
            waiter.run(retries=1)

        assert waiter.get_result() == example_response

    def test_condition_not_met_without_error(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        example_response: Response,
    ):
        waiter = RequestsWaiter(request=prepared_request, session=session_mock)
        waiter.add_checker(expected_value=False, comparer="is_equal", dict_path="ok")
        waiter.run(retries=1, raise_error=False)

        assert waiter.get_result() == example_response

    def test_wait_for_value(
        self,
        mocker: MockerFixture,
        prepared_request: PreparedRequest,
        example_response: Response,
        caplog: LogCaptureFixture,
    ):
        res = Response()
        res.status_code = 200
        res.request = prepared_request
        mocker.patch("uuid.uuid4", side_effect=["TEST1", "TEST2", "TEST3", "TEST4"])
        mocker.patch(
            "requests.Session.send",
            side_effect=[res, example_response],
        )
        logs = [
            (
                "bepatient.waiter_src.executors.requests_executor",
                10,
                "Creating a new Session object",
            ),
            (
                "bepatient.waiter_src.waiter",
                20,
                "Checking whether the condition has been met. The 1 approach",
            ),
            (
                "bepatient.waiter_src.executors.requests_executor",
                20,
                "Sent: curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123'"
                " https://webludus.pl/",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: TEST1 | Checker: StatusCodeChecker | Comparer: is_equal"
                " | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: TEST1 | Response status code: 200"
                " | Response content: None",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                10,
                "Check success! | uuid: TEST1 | Checker: StatusCodeChecker"
                " | Comparer: is_equal | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: TEST2 | Checker: HeadersChecker | Comparer: is_equal"
                " | Dictor_fallback: None | Expected_value: WebLudus.pl"
                " | Ignore_case: False | Path: Server | Search_query: None"
                " | Data: WebLudus.pl",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: TEST2 | Response headers: {}",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: TEST2 | Dictor path: Server | Dictor search: None"
                " | Dictor data: None",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: TEST2 | Condition not met | Checker: HeadersChecker"
                " | Comparer: is_equal | Dictor_fallback: None"
                " | Expected_value: WebLudus.pl | Ignore_case: False | Path: Server"
                " | Search_query: None | Data: WebLudus.pl",
            ),
            (
                "bepatient.waiter_src.waiter",
                20,
                "The condition has not been met. Waiting time: 1",
            ),
            (
                "bepatient.waiter_src.waiter",
                20,
                "Checking whether the condition has been met. The 2 approach",
            ),
            (
                "bepatient.waiter_src.executors.requests_executor",
                20,
                "Sent: curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123' "
                "https://webludus.pl/",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: TEST3 | Checker: StatusCodeChecker | Comparer: is_equal"
                " | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: TEST3 | Response status code: 200 | Response content:"
                ' b\'{"list_of_dicts": [{"name": "John", "age": 30}, {"name": "Mike",'
                ' "age": 15}], "ok": true, "some_number": 123, "list": ["1", "2", "3"],'
                ' "none": null, "empty": "", "false": false, "name": "Jack",'
                ' "City": "Cracow"}\'',
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                10,
                "Check success! | uuid: TEST3 | Checker: StatusCodeChecker | Comparer: "
                "is_equal | Expected_value: 200 | Data: 200",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                20,
                "Check uuid: TEST4 | Checker: HeadersChecker | Comparer: is_equal"
                " | Dictor_fallback: None | Expected_value: WebLudus.pl"
                " | Ignore_case: False | Path: Server | Search_query: None"
                " | Data: WebLudus.pl",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: TEST4 | Response headers: {'Content-Language': 'en-US',"
                " 'Content-Type': 'application/json', 'Server': 'WebLudus.pl',"
                " 'X-Render-Origin_Server': 'gunicorn'}",
            ),
            (
                "bepatient.waiter_src.checkers.response_checkers",
                20,
                "Check uuid: TEST4 | Dictor path: Server | Dictor search: None"
                " | Dictor data: WebLudus.pl",
            ),
            (
                "bepatient.waiter_src.checkers.checker",
                10,
                "Check success! | uuid: TEST4 | Checker: HeadersChecker"
                " | Comparer: is_equal | Dictor_fallback: None"
                " | Expected_value: WebLudus.pl | Ignore_case: False | Path: Server"
                " | Search_query: None | Data: WebLudus.pl",
            ),
            ("bepatient.waiter_src.waiter", 20, "Condition met!"),
        ]

        waiter = RequestsWaiter(request=prepared_request)
        waiter.add_checker(
            expected_value="WebLudus.pl",
            comparer="is_equal",
            checker="headers_checker",
            dict_path="Server",
        )
        waiter.run(retries=2)

        assert waiter.get_result() == example_response
        assert caplog.record_tuples == logs


class TestWaitForValueInRequests:
    def test_happy_path(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        example_response: Response,
    ):
        result = wait_for_value_in_request(
            request=prepared_request,
            session=session_mock,
            comparer="is_equal",
            expected_value=True,
            dict_path="ok",
            retries=1,
        )

        assert result == example_response

    def test_condition_not_met_raise_error(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        error_msg: str,
    ):
        with pytest.raises(WaiterConditionWasNotMet, match=error_msg):
            wait_for_value_in_request(
                request=prepared_request,
                session=session_mock,
                comparer="is_equal",
                expected_value=False,
                dict_path="ok",
                retries=1,
            )


class TestWaitForValuesInRequests:
    def test_happy_path(
        self,
        prepared_request: PreparedRequest,
        session_mock: Session,
        example_response: Response,
    ):
        list_of_checkers = [
            {
                "checker": "json_checker",
                "comparer": "contain",
                "expected_value": "John",
                "dict_path": "list_of_dicts",
                "search_query": "name",
            },
            {
                "checker": "headers_checker",
                "comparer": "is_equal",
                "expected_value": "en-US",
                "dict_path": "Content-Language",
            },
        ]

        response = wait_for_values_in_request(
            request=prepared_request,
            session=session_mock,
            checkers=list_of_checkers,
            retries=1,
        )

        assert response == example_response

    def test_condition_not_met_raise_error(
        self, prepared_request: PreparedRequest, session_mock: Session
    ):
        msg = (
            "The condition has not been met! | Failed checkers:"
            " (Checker: JsonChecker | Comparer: contain | Dictor_fallback: None"
            " | Expected_value: Jerry | Path: list_of_dicts | Search_query: name"
            " | Data: ['John', 'Mike'], Checker: HeadersChecker | Comparer: is_equal"
            " | Dictor_fallback: None | Expected_value: xml | Path: content"
            " | Search_query: None | Data: json) | curl -X GET -H 'task: test' -H"
            " 'Cookie: user-token=abc-123' https://webludus.pl/"
        )
        list_of_checkers = [
            {
                "checker": "json_checker",
                "comparer": "contain",
                "expected_value": "Jerry",
                "dict_path": "list_of_dicts",
                "search_query": "name",
            },
            {
                "checker": "headers_checker",
                "comparer": "is_equal",
                "expected_value": "xml",
                "dict_path": "content",
            },
        ]

        with pytest.raises(WaiterConditionWasNotMet, match=msg):
            wait_for_values_in_request(
                request=prepared_request,
                session=session_mock,
                checkers=list_of_checkers,
                retries=1,
            )


def test_dict_differences():
    expected_dict = {
        "Key1": 1,
        "Key2": None,
        "Key3": "STRING",
        "Key4": "Another string",
    }
    actual_dict = {"Key1": 2, "Key5": False, "Key3": "STRING", "1": True}
    msg = {
        "key_missing_exp": {"Key2", "Key4"},
        "key_missing_actual": {"Key5", "1"},
        "value_diff": {"Key1": {"expected": 1, "actual": 2}},
    }
    assert dict_differences(expected_dict=expected_dict, actual_dict=actual_dict) == msg


def test_to_curl(prepared_request: PreparedRequest):
    expected_curl = (
        "curl -X GET -H 'task: test' -H 'Cookie: user-token=abc-123' "
        "https://webludus.pl/"
    )

    assert to_curl(prepared_request) == expected_curl



import logging

from requests import PreparedRequest, Request, Response, Session
from requests.exceptions import RequestException

from bepatient.curler import Curler
from bepatient.waiter_src.checkers.response_checkers import StatusCodeChecker
from bepatient.waiter_src.comparators import is_equal

from .executor import Executor

log = logging.getLogger(__name__)


# pylint: disable-next=too-many-instance-attributes
class RequestsExecutor(Executor):
    """An executor that sends a request and waits for a certain condition to be met.
    Args:
        req_or_res (PreparedRequest | Request | Response): request to send.
        expected_status_code (int): expected HTTP status code of the response
        session (Session | None, optional): requests session to use.
        timeout (int | tuple[int, int] | None, optional): request timeout in seconds.
            Default value is 15 for connect and 30 for read (15, 30). If user provide
            one value, it will be applied to both - connect and read timeouts."""

    def __init__(
        self,
        req_or_res: PreparedRequest | Request | Response,
        expected_status_code: int,
        session: Session | None = None,
        timeout: int | tuple[int, int] | None = None,
    ):
        super().__init__()
        self._result: Response | None = None
        self._take_from_result: bool = False
        self._status_code_checker = StatusCodeChecker(is_equal, expected_status_code)

        if timeout:
            self.timeout = timeout
        else:
            self.timeout = (15, 30)

        if session:
            self.session = session
        else:
            log.debug("Creating a new Session object")
            self.session = Session()

        if isinstance(req_or_res, Request):
            self.request = self.session.prepare_request(req_or_res)
        elif isinstance(req_or_res, PreparedRequest):
            self.request = req_or_res
        else:
            self._result = req_or_res
            self._take_from_result = True
            if len(self._result.history) > 0:
                self.request = self._result.history[0].request
            else:
                self.request = self._result.request
            self._merge_session_data_to_prepared_request()

        self._input = Curler().to_curl(self.request)

    def _merge_session_data_to_prepared_request(self):
        log.debug("Merging session.headers into PreparedRequest object")
        self.request.headers.update(self.session.headers)  # type: ignore
        if self.session.cookies:
            log.debug("Merging session.cookies into PreparedRequest object")
            req_cookies = self.request.headers.get("Cookie", "")
            if req_cookies:
                log.debug("PreparedRequest already has cookies")
                req_cookies = req_cookies + "; "
            session_cookies = "; ".join(
                f"{k}={v}" for k, v in self.session.cookies.items()
            )
            self.request.headers["Cookie"] = req_cookies + session_cookies

    def is_condition_met(self) -> bool:
        """Sends the request and check if all checkers pass or timeout occurs.

        Returns:
            bool: True if all checkers pass, False otherwise.

        Raises:
            ExecutorIsNotReady: If the executor is not ready to send the request."""
        if not self._take_from_result:
            try:
                self._result = self.session.send(
                    request=self.request, timeout=self.timeout
                )
                self._input = Curler().to_curl(self._result)
                log.info("Sent: %s", self._input)
            except RequestException:
                log.exception("RequestException! CURL: %s", self._input)
                return False
        else:
            self._take_from_result = False

        if self._status_code_checker.check(self._result):
            self._failed_checkers = [
                checker for checker in self._checkers if not checker.check(self._result)
            ]
        else:
            self._failed_checkers = [self._status_code_checker]

        if len(self._failed_checkers) == 0:
            return True
        return False
